@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2022 Nicotine+ Contributors
+# COPYRIGHT (C) 2020-2023 Nicotine+ Contributors
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2016-2018 Mutnick <mutnick@techie.com>
 # COPYRIGHT (C) 2008-2011 quinox <quinox@users.sf.net>
@@ -24,10 +24,6 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-"""
-This module contains configuration classes for Nicotine.
-"""
-
 import configparser
 import os
 import sys
@@ -35,10 +31,15 @@ import sys
 from ast import literal_eval
 from collections import defaultdict
 
+from pynicotine.events import events
+from pynicotine.i18n import apply_translations
+from pynicotine.utils import encode_path
+from pynicotine.utils import load_file
+from pynicotine.utils import write_file_and_backup
+
 
 class Config:
-    """
-    This class holds configuration information and provides the
+    """This class holds configuration information and provides the
     following methods:
 
     need_config() - returns true if configuration information is incomplete
@@ -52,26 +53,9 @@ class Config:
 
     def __init__(self):
 
-        config_dir, self.data_dir = self.get_user_directories()
-        self.filename = os.path.join(config_dir, "config")
-        self.plugin_dir = os.path.join(self.data_dir, "plugins")
-        self.version = "3.2.7"
-        self.python_version = sys.version
-        self.gtk_version = ""
-
-        self.application_name = "Nicotine+"
-        self.application_id = "org.nicotine_plus.Nicotine"
-        self.summary = _("Graphical client for the Soulseek peer-to-peer network")
-        self.author = "Nicotine+ Team"
-        self.copyright = """© 2004–2022 Nicotine+ Contributors
-© 2003–2004 Nicotine Contributors
-© 2001–2003 PySoulSeek Contributors"""
-
-        self.website_url = "https://nicotine-plus.org/"
-        self.privileges_url = "https://www.slsknet.org/userlogin.php?username=%s"
-        self.portchecker_url = "http://tools.slsknet.org/porttest.php?port=%s"
-        self.issue_tracker_url = "https://github.com/nicotine-plus/nicotine-plus/issues"
-        self.translations_url = "https://nicotine-plus.org/doc/TRANSLATIONS"
+        config_folder_path, data_folder_path = self.get_user_folders()
+        self.set_config_file(os.path.join(config_folder_path, "config"))
+        self.set_data_folder(data_folder_path)
 
         self.config_loaded = False
         self.parser = configparser.ConfigParser(strict=False, interpolation=None)
@@ -80,87 +64,89 @@ class Config:
         self.removed_options = {}
 
     @staticmethod
-    def get_user_directories():
-        """ Returns a tuple:
-        - the config directory
-        - the data directory """
+    def get_user_folders():
+        """Returns a tuple:
+
+        - the config folder
+        - the data folder
+        """
 
         if sys.platform == "win32":
             try:
-                data_dir = os.path.join(os.environ['APPDATA'], 'nicotine')
+                data_folder_path = os.path.join(os.path.normpath(os.environ["APPDATA"]), "nicotine")
             except KeyError:
-                data_dir, _filename = os.path.split(sys.argv[0])
+                data_folder_path, _basename = os.path.split(sys.argv[0])
 
-            config_dir = os.path.join(data_dir, "config")
-            return config_dir, data_dir
+            config_folder_path = os.path.join(data_folder_path, "config")
+            return config_folder_path, data_folder_path
 
         home = os.path.expanduser("~")
+        legacy_folder_path = os.path.join(home, ".nicotine")
 
-        legacy_dir = os.path.join(home, '.nicotine')
-
-        if os.path.isdir(legacy_dir.encode("utf-8")):
-            return legacy_dir, legacy_dir
+        if os.path.isdir(legacy_folder_path.encode("utf-8")):
+            return legacy_folder_path, legacy_folder_path
 
         def xdg_path(xdg, default):
             path = os.environ.get(xdg)
+            path = path.split(":")[0] if path else default
 
-            path = path.split(':')[0] if path else default
+            return os.path.join(path, "nicotine")
 
-            return os.path.join(path, 'nicotine')
+        config_folder_path = xdg_path("XDG_CONFIG_HOME", os.path.join(home, ".config"))
+        data_folder_path = xdg_path("XDG_DATA_HOME", os.path.join(home, ".local", "share"))
 
-        config_dir = xdg_path('XDG_CONFIG_HOME', os.path.join(home, '.config'))
-        data_dir = xdg_path('XDG_DATA_HOME', os.path.join(home, '.local', 'share'))
+        return config_folder_path, data_folder_path
 
-        return config_dir, data_dir
+    def set_config_file(self, file_path):
+        self.config_file_path = os.path.abspath(file_path)
+
+    def set_data_folder(self, folder_path):
+        self.data_folder_path = os.environ["NICOTINE_DATA_HOME"] = os.path.abspath(folder_path)
 
     def create_config_folder(self):
-        """ Create the folder for storing the config file in, if the folder
-        doesn't exist """
+        """Create the folder for storing the config file in, if the folder
+        doesn't exist."""
 
-        path, _filename = os.path.split(self.filename)
+        folder_path, _basename = os.path.split(self.config_file_path)
 
-        if not path:
+        if not folder_path:
             # Only file name specified, use current folder
             return True
 
-        from pynicotine.utils import encode_path
-        path_encoded = encode_path(path)
+        folder_path_encoded = encode_path(folder_path)
 
         try:
-            if not os.path.isdir(path_encoded):
-                os.makedirs(path_encoded)
+            if not os.path.isdir(folder_path_encoded):
+                os.makedirs(folder_path_encoded)
 
-        except OSError as msg:
+        except OSError as error:
             from pynicotine.logfacility import log
 
             log.add(_("Can't create directory '%(path)s', reported error: %(error)s"),
-                    {'path': path, 'error': msg})
+                    {"path": folder_path, "error": error})
             return False
 
         return True
 
     def create_data_folder(self):
-        """ Create the folder for storing data in (aliases, shared files etc.),
-        if the folder doesn't exist """
+        """Create the folder for storing data in (shared files etc.), if the
+        folder doesn't exist."""
 
-        from pynicotine.utils import encode_path
-        data_dir_encoded = encode_path(self.data_dir)
+        data_folder_path_encoded = encode_path(self.data_folder_path)
 
         try:
-            if not os.path.isdir(data_dir_encoded):
-                os.makedirs(data_dir_encoded)
+            if not os.path.isdir(data_folder_path_encoded):
+                os.makedirs(data_folder_path_encoded)
 
-        except OSError as msg:
+        except OSError as error:
             from pynicotine.logfacility import log
 
             log.add(_("Can't create directory '%(path)s', reported error: %(error)s"),
-                    {'path': self.data_dir, 'error': msg})
+                    {"path": self.data_folder_path, "error": error})
 
     def load_config(self):
 
-        from pynicotine.utils import load_file
-
-        log_dir = os.path.join(self.data_dir, "logs")
+        log_folder_path = os.path.join("${NICOTINE_DATA_HOME}", "logs")
         self.defaults = {
             "server": {
                 "server": ("server.slsknet.org", 2242),
@@ -170,7 +156,7 @@ class Config:
                 "ctcpmsgs": False,
                 "autosearch": [],
                 "autoreply": "",
-                "portrange": (2234, 2239),
+                "portrange": (2234, 2234),
                 "upnp": True,
                 "upnp_interval": 4,
                 "auto_connect_startup": True,
@@ -186,26 +172,25 @@ class Config:
                 "command_aliases": {}
             },
             "transfers": {
-                "incompletedir": os.path.join(self.data_dir, 'incomplete'),
-                "downloaddir": os.path.join(self.data_dir, 'downloads'),
-                "uploaddir": os.path.join(self.data_dir, 'received'),
+                "incompletedir": os.path.join("${NICOTINE_DATA_HOME}", "incomplete"),
+                "downloaddir": os.path.join("${NICOTINE_DATA_HOME}", "downloads"),
+                "uploaddir": os.path.join("${NICOTINE_DATA_HOME}", "received"),
                 "usernamesubfolders": False,
                 "shared": [],
                 "buddyshared": [],
+                "trustedshared": [],
                 "uploadbandwidth": 50,
-                "uselimit": False,
-                "usealtlimits": False,
+                "use_upload_speed_limit": "unlimited",
                 "uploadlimit": 1000,
                 "uploadlimitalt": 100,
-                "downloadlimit": 0,
+                "use_download_speed_limit": "unlimited",
+                "downloadlimit": 1000,
                 "downloadlimitalt": 100,
                 "preferfriends": False,
                 "useupslots": False,
                 "uploadslots": 2,
                 "afterfinish": "",
                 "afterfolder": "",
-                "lock": True,
-                "reverseorder": False,
                 "fifoqueue": False,
                 "usecustomban": False,
                 "limitby": True,
@@ -214,23 +199,22 @@ class Config:
                 "customgeoblock": "Sorry, your country is blocked",
                 "queuelimit": 10000,
                 "filelimit": 100,
-                "buddysharestrustedonly": False,
+                "reveal_buddy_shares": False,
+                "reveal_trusted_shares": False,
                 "friendsnolimits": False,
                 "groupdownloads": "folder_grouping",
                 "groupuploads": "folder_grouping",
                 "geoblock": False,
                 "geoblockcc": [""],
-                "remotedownloads": True,
-                "uploadallowed": 2,
+                "remotedownloads": False,
+                "uploadallowed": 3,
                 "autoclear_downloads": False,
                 "autoclear_uploads": False,
-                "uploadsinsubdirs": True,
                 "rescanonstartup": True,
                 "enablefilters": False,
                 "downloadregexp": "",
                 "downloadfilters": [
                     ["desktop.ini", 1],
-                    ["folder.jpg", 1],
                     ["*.url", 1],
                     ["thumbs.db", 1],
                     ["albumart(_{........-....-....-....-............}_)?(_?(large|small))?\\.jpg", 0]
@@ -240,13 +224,13 @@ class Config:
                 "downloadsexpanded": True,
                 "uploadsexpanded": True
             },
+            "userbrowse": {
+                "expand_folders": True
+            },
             "userinfo": {
                 "descr": "''",
                 "pic": ""
             },
-            "userbrowse": {
-                "expand_folders": True
-            },
             "words": {
                 "censored": [],
                 "autoreplaced": {
@@ -258,38 +242,35 @@ class Config:
                     "thier": "their",
                     "tihs": "this"
                 },
-                "censorfill": "*",
                 "censorwords": False,
                 "replacewords": False,
                 "tab": True,
-                "cycle": False,
                 "dropdown": False,
                 "characters": 3,
                 "roomnames": False,
                 "buddies": True,
                 "roomusers": True,
-                "commands": True,
-                "aliases": True,
-                "onematch": False
+                "commands": True
             },
             "logging": {
                 "debug": False,
                 "debugmodes": [],
-                "debuglogsdir": os.path.join(log_dir, "debug"),
+                "debuglogsdir": os.path.join(log_folder_path, "debug"),
                 "logcollapsed": True,
-                "transferslogsdir": os.path.join(log_dir, "transfers"),
-                "rooms_timestamp": "%H:%M:%S",
-                "private_timestamp": "%Y-%m-%d %H:%M:%S",
-                "log_timestamp": "%Y-%m-%d %H:%M:%S",
+                "transferslogsdir": os.path.join(log_folder_path, "transfers"),
+                "rooms_timestamp": "%X",
+                "private_timestamp": "%x %X",
+                "log_timestamp": "%x %X",
                 "privatechat": True,
                 "chatrooms": True,
                 "transfers": False,
                 "debug_file_output": False,
-                "roomlogsdir": os.path.join(log_dir, "rooms"),
-                "privatelogsdir": os.path.join(log_dir, "private"),
+                "roomlogsdir": os.path.join(log_folder_path, "rooms"),
+                "privatelogsdir": os.path.join(log_folder_path, "private"),
                 "readroomlogs": True,
-                "readroomlines": 15,
-                "readprivatelines": 15,
+                "readroomlines": 200,
+                "readprivatelines": 200,
+                "private_chats": [],
                 "rooms": []
             },
             "privatechat": {
@@ -312,31 +293,33 @@ class Config:
                 "history": [],
                 "enablefilters": False,
                 "filters_visible": False,
-                "defilter": ["", "", "", "", 0, ""],
+                "defilter": ["", "", "", "", 0, "", ""],
                 "filtercc": [],
                 "filterin": [],
                 "filterout": [],
                 "filtersize": [],
                 "filterbr": [],
                 "filtertype": [],
+                "filterlength": [],
                 "search_results": True,
                 "max_displayed_results": 1500,
                 "min_search_chars": 3,
-                "remove_special_chars": True,
                 "private_search_results": True
             },
             "ui": {
+                "language": "",
                 "dark_mode": False,
                 "header_bar": True,
                 "icontheme": "",
-                "chatme": "#908e8b",
+                "chatme": "#908E8B",
                 "chatremote": "",
                 "chatlocal": "",
-                "chathilite": "#5288ce",
-                "urlcolor": "#5288ce",
-                "useronline": "#16bb5c",
-                "useraway": "#c9ae13",
-                "useroffline": "#e04f5e",
+                "chatcommand": "#908E8B",
+                "chathilite": "#5288CE",
+                "urlcolor": "#5288CE",
+                "useronline": "#16BB5C",
+                "useraway": "#C9AE13",
+                "useroffline": "#E04F5E",
                 "usernamehotspots": True,
                 "usernamestyle": "bold",
                 "textbg": "",
@@ -346,8 +329,8 @@ class Config:
                 "spellcheck": True,
                 "exitdialog": 1,
                 "tab_default": "",
-                "tab_hilite": "#497ec2",
-                "tab_changed": "#497ec2",
+                "tab_hilite": "#497EC2",
+                "tab_changed": "#497EC2",
                 "tab_select_previous": True,
                 "tabmain": "Top",
                 "tabrooms": "Top",
@@ -355,8 +338,8 @@ class Config:
                 "tabinfo": "Top",
                 "tabbrowse": "Top",
                 "tabsearch": "Top",
-                "tab_status_icons": True,
                 "globalfont": "",
+                "textviewfont": "",
                 "chatfont": "",
                 "tabclosers": True,
                 "searchfont": "",
@@ -398,9 +381,8 @@ class Config:
                 "xposition": -1,
                 "yposition": -1,
                 "maximized": True,
-                "urgencyhint": True,
-                "file_path_tooltips": True,
-                "reverse_file_paths": True
+                "reverse_file_paths": True,
+                "file_size_unit": ""
             },
             "private_rooms": {
                 "rooms": {}
@@ -427,13 +409,15 @@ class Config:
                 "notification_popup_folder": True,
                 "notification_popup_private_message": True,
                 "notification_popup_chatroom": False,
-                "notification_popup_chatroom_mention": True
+                "notification_popup_chatroom_mention": True,
+                "notification_popup_wish": True
             },
             "plugins": {
                 "enable": True,
                 "enabled": []
             },
             "statistics": {
+                "since_timestamp": 0,
                 "started_downloads": 0,
                 "completed_downloads": 0,
                 "downloaded_size": 0,
@@ -454,7 +438,13 @@ class Config:
                 "geopanic",
                 "enablebuddyshares",
                 "friendsonly",
-                "enabletransferbuttons"
+                "enabletransferbuttons",
+                "uselimit",
+                "usealtlimits",
+                "uploadsinsubdirs",
+                "reverseorder",
+                "lock",
+                "buddysharestrustedonly"
             ),
             "server": (
                 "lastportstatuscheck",
@@ -492,7 +482,10 @@ class Config:
                 "notexists",
                 "roomlistcollapsed",
                 "showaway",
-                "decimalsep"
+                "decimalsep",
+                "urgencyhint",
+                "tab_status_icons",
+                "file_path_tooltips"
             ),
             "columns": (
                 "downloads",
@@ -522,7 +515,8 @@ class Config:
                 "distrib_ignore",
                 "reopen_tabs",
                 "max_stored_results",
-                "re_filter"
+                "re_filter",
+                "remove_special_chars"
             ),
             "userinfo": (
                 "descrutf8"
@@ -549,97 +543,44 @@ class Config:
             ),
             "notifications": (
                 "notification_tab_icons"
+            ),
+            "words": (
+                "cycle",
+                "onematch",
+                "aliases",
+                "censorfill"
+            ),
+            "players": (
+                "default"
             )
         }
 
-        # Initialize config with default values
-        for key, value in self.defaults.items():
-            self.sections[key] = value.copy()
-
         self.create_config_folder()
         self.create_data_folder()
 
-        load_file(self.filename, self.parse_config)
+        load_file(self.config_file_path, self.parse_config)
 
         # Update config values from file
         self.set_config()
 
-        if sys.platform == "darwin":
-            # Disable header bar in macOS for now due to GTK 3 performance issues
-            self.sections["ui"]["header_bar"] = False
+        language = self.sections["ui"]["language"]
 
-        # Convert special download folder share to regular share
-        if self.sections["transfers"].get("sharedownloaddir", False):
-            shares = self.sections["transfers"]["shared"]
-            virtual_name = "Downloaded"
-            shared_folder = (virtual_name, self.sections["transfers"]["downloaddir"])
-
-            if shared_folder not in shares and virtual_name not in (x[0] for x in shares):
-                shares.append(shared_folder)
-
-        # Load command aliases from legacy file
-        try:
-            from pynicotine.utils import encode_path
-            encoded_alias_path = encode_path(self.filename + ".alias")
-
-            if not self.sections["server"]["command_aliases"] and os.path.exists(encoded_alias_path):
-                with open(encoded_alias_path, 'rb') as file_handle:
-                    from pynicotine.utils import RestrictedUnpickler
-
-                    self.sections["server"]["command_aliases"] = RestrictedUnpickler(
-                        file_handle, encoding='utf-8').load()
-
-        except Exception:
-            pass
-
-        self.config_loaded = True
+        if language:
+            apply_translations(language)
 
         from pynicotine.logfacility import log
-        log.add_debug("Using configuration: %(file)s", {"file": self.filename})
+        log.init_log_levels()
+        log.update_folder_paths()
+        log.add_debug("Using configuration: %(file)s", {"file": self.config_file_path})
 
-    def parse_config(self, filename):
-        """ Parses the config file """
+        events.connect("quit", self._quit)
 
-        from pynicotine.utils import encode_path
+    def parse_config(self, file_path):
+        """Parses the config file."""
 
-        try:
-            with open(encode_path(filename), 'a+', encoding="utf-8") as file_handle:
-                file_handle.seek(0)
-                self.parser.read_file(file_handle)
-
-        except UnicodeDecodeError:
-            self.convert_config()
-            self.parse_config(filename)
-
-    def convert_config(self):
-        """ Converts the config to utf-8.
-        Mainly for upgrading Windows build. (22 July, 2020) """
-
-        try:
-            from chardet import detect
-
-        except ImportError:
-            from pynicotine.logfacility import log
-
-            log.add("Failed to convert config file to UTF-8. Please install python3-chardet and start "
-                    "the application again.")
-            sys.exit()
-
-        from pynicotine.utils import encode_path
-        conv_filename = encode_path(self.filename + ".conv")
-        os.replace(self.filename, conv_filename)
-
-        with open(conv_filename, 'rb') as file_handle:
-            rawdata = file_handle.read()
-
-        from_encoding = detect(rawdata)['encoding']
-
-        with open(conv_filename, encoding=from_encoding) as file_read:
-            with open(encode_path(self.filename), 'w', encoding="utf-8") as file_write:
-                for line in file_read:
-                    file_write.write(line[:-1] + '\r\n')
-
-        os.remove(conv_filename)
+        with open(encode_path(file_path), "a+", encoding="utf-8") as file_handle:
+            file_handle.seek(0)
+            self.parser.read_file(file_handle)
 
     def need_config(self):
 
@@ -650,7 +591,7 @@ class Config:
         return False
 
     def set_config(self):
-        """ Set config values parsed from file earlier """
+        """Set config values parsed from file earlier."""
 
         from pynicotine.logfacility import log
 
@@ -659,17 +600,17 @@ class Config:
 
                 # Check if config section exists in defaults
                 if i not in self.defaults and i not in self.removed_options:
-                    log.add(_("Unknown config section '%s'"), i)
+                    log.add_debug("Unknown config section '%s'", i)
 
                 # Check if config option exists in defaults
                 elif (j not in self.defaults.get(i, {}) and j not in self.removed_options.get(i, {})
                         and i != "plugins" and j != "filter"):
-                    log.add(_("Unknown config option '%(option)s' in section '%(section)s'"),
-                            {'option': j, 'section': i})
+                    log.add_debug("Unknown config option '%(option)s' in section '%(section)s'",
+                                  {"option": j, "section": i})
 
                 else:
-                    """ Attempt to get the default value for a config option. If there's no default
-                    value, it's a custom option from a plugin, so no checks are needed. """
+                    # Attempt to get the default value for a config option. If there's no default
+                    # value, it's a custom option from a plugin, so no checks are needed.
 
                     try:
                         default_val = self.defaults[i][j]
@@ -683,8 +624,8 @@ class Config:
                         self.sections[i][j] = val
                         continue
 
-                    """ Check that the value of a config option is of the same type as the default
-                    value. If that's not the case, reset the value. """
+                    # Check that the value of a config option is of the same type as the default
+                    # value. If that's not the case, reset the value.
 
                     try:
                         if not isinstance(default_val, str):
@@ -698,7 +639,7 @@ class Config:
                         if i != "plugins" and j != "filter":
                             if (isinstance(eval_val, type(default_val))
                                     or (isinstance(default_val, bool)
-                                        and isinstance(eval_val, int) and eval_val in (0, 1))):
+                                        and isinstance(eval_val, int) and eval_val in {0, 1})):
                                 # Value is valid
                                 pass
 
@@ -712,32 +653,93 @@ class Config:
                         self.sections[i][j] = default_val
 
                         log.add("Config error: Couldn't decode '%s' section '%s' value '%s', value has been reset", (
-                            (i[:120] + '…') if len(i) > 120 else i,
-                            (j[:120] + '…') if len(j) > 120 else j,
-                            (val[:120] + '…') if len(val) > 120 else val
+                            (i[:120] + "…") if len(i) > 120 else i,
+                            (j[:120] + "…") if len(j) > 120 else j,
+                            (val[:120] + "…") if len(val) > 120 else val
                         ))
 
-        server = self.sections["server"]
+        # Add any default options not present in the config file
+        for section, options in self.defaults.items():
+            if section not in self.sections:
+                self.sections[section] = {}
+
+            for option, value in options.items():
+                if option in self.sections[section]:
+                    continue
+
+                # Migrate download speed limit preference (3.3.0)
+                if option == "use_download_speed_limit" and section == "transfers":
+                    if self.sections[section].get("usealtlimits", False):
+                        use_speed_limit = "alternative"
+
+                    elif self.sections[section].get("downloadlimit", 0) > 0:
+                        use_speed_limit = "primary"
+
+                    else:
+                        use_speed_limit = "unlimited"
+
+                    self.sections[section][option] = use_speed_limit
+                    continue
+
+                # Migrate upload speed limit preference (3.3.0)
+                if option == "use_upload_speed_limit" and section == "transfers":
+                    if self.sections[section].get("usealtlimits", False):
+                        use_speed_limit = "alternative"
+
+                    elif self.sections[section].get("uselimit", False):
+                        use_speed_limit = "primary"
+
+                    else:
+                        use_speed_limit = "unlimited"
+
+                    self.sections[section][option] = use_speed_limit
+                    continue
+
+                # Set default value
+                self.sections[section][option] = value
+
+        # Convert special download folder share to regular share
+        if self.sections["transfers"].get("sharedownloaddir", False):
+            shares = self.sections["transfers"]["shared"]
+            virtual_name = "Downloaded"
+            shared_folder = (virtual_name, self.sections["transfers"]["downloaddir"])
+
+            if shared_folder not in shares and virtual_name not in (x[0] for x in shares):
+                shares.append(shared_folder)
+
+        # Migrate old trusted buddy shares to new format (3.3.0)
+        if self.sections["transfers"].get("buddysharestrustedonly", False):
+            buddy_shares = self.sections["transfers"]["buddyshared"]
+
+            self.sections["transfers"]["trustedshared"] = buddy_shares[:]
+            buddy_shares.clear()
+
+        # Migrate old media player command to new format (3.3.0)
+        old_default_player = self.sections["players"].get("default", None)
+
+        if old_default_player:
+            self.sections["urls"]["protocols"]["audio"] = old_default_player
+
+        # Enable previously disabled header bar on macOS (3.3.0)
+        if sys.platform == "darwin" and old_default_player is not None:
+            self.sections["ui"]["header_bar"] = True
 
         # Check if server value is valid
-        if (len(server["server"]) != 2
-                or not isinstance(server["server"][0], str)
-                or not isinstance(server["server"][1], int)):
+        server_addr = self.sections["server"]["server"]
 
-            server["server"] = self.defaults["server"]["server"]
+        if (len(server_addr) != 2 or not isinstance(server_addr[0], str) or not isinstance(server_addr[1], int)):
+            self.sections["server"]["server"] = self.defaults["server"]["server"]
 
         # Check if port range value is valid
-        if (len(server["portrange"]) != 2
-                or not all(isinstance(i, int) for i in server["portrange"])):
+        port_range = self.sections["server"]["portrange"]
 
-            server["portrange"] = self.defaults["server"]["portrange"]
+        if (len(port_range) != 2 or not all(isinstance(i, int) for i in port_range)):
+            self.sections["server"]["portrange"] = self.defaults["server"]["portrange"]
 
-        else:
-            # Setting the port range in numerical order
-            server["portrange"] = (min(server["portrange"]), max(server["portrange"]))
+        self.config_loaded = True
 
-    def write_config_callback(self, filename):
-        self.parser.write(filename)
+    def write_config_callback(self, file_path):
+        self.parser.write(file_path)
 
     def write_configuration(self):
 
@@ -767,37 +769,44 @@ class Config:
             return
 
         from pynicotine.logfacility import log
-        from pynicotine.utils import write_file_and_backup
 
-        write_file_and_backup(self.filename, self.write_config_callback, protect=True)
-        log.add_debug("Saved configuration: %(file)s", {"file": self.filename})
+        write_file_and_backup(self.config_file_path, self.write_config_callback, protect=True)
+        log.add_debug("Saved configuration: %(file)s", {"file": self.config_file_path})
 
-    def write_config_backup(self, filename):
+    def write_config_backup(self, file_path):
 
         from pynicotine.logfacility import log
-        from pynicotine.utils import encode_path
 
-        if not filename.endswith(".tar.bz2"):
-            filename += ".tar.bz2"
+        if not file_path.endswith(".tar.bz2"):
+            file_path += ".tar.bz2"
 
-        filename_encoded = encode_path(filename)
+        file_path_encoded = encode_path(file_path)
 
         try:
-            if os.path.exists(filename_encoded):
-                raise FileExistsError("File %s exists" % filename)
+            if os.path.exists(file_path_encoded):
+                raise FileExistsError(f"File {file_path} exists")
 
             import tarfile
-            with tarfile.open(filename_encoded, "w:bz2") as tar:
-                if not os.path.exists(filename_encoded):
+            with tarfile.open(file_path_encoded, "w:bz2") as tar:
+                if not os.path.exists(file_path_encoded):
                     raise FileNotFoundError("Config file missing")
 
-                tar.add(self.filename)
+                tar.add(self.config_file_path)
 
         except Exception as error:
             log.add(_("Error backing up config: %s"), error)
             return
 
-        log.add(_("Config backed up to: %s"), filename)
+        log.add(_("Config backed up to: %s"), file_path)
+
+    def _quit(self):
+
+        self.parser.clear()
+        self.sections.clear()
+        self.defaults.clear()
+        self.removed_options.clear()
+
+        self.config_loaded = False
 
 
 config = Config()
