@@ -1,7 +1,7 @@
-# COPYRIGHT (C) 2020-2022 Nicotine+ Contributors
+# COPYRIGHT (C) 2020-2024 Nicotine+ Contributors
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
-# COPYRIGHT (C) 2008-2009 Quinox <quinox@users.sf.net>
-# COPYRIGHT (C) 2006-2009 Daelstorm <daelstorm@gmail.com>
+# COPYRIGHT (C) 2008-2009 quinox <quinox@users.sf.net>
+# COPYRIGHT (C) 2006-2009 daelstorm <daelstorm@gmail.com>
 # COPYRIGHT (C) 2003-2004 Hyriand <hyriand@thegraveyard.org>
 #
 # GNU GENERAL PUBLIC LICENSE
@@ -20,32 +20,28 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-import random
-import string
-
 from gi.repository import Gdk
 from gi.repository import Gio
 from gi.repository import GLib
 from gi.repository import Gtk
 
-from pynicotine import slskmessages
 from pynicotine.config import config
+from pynicotine.core import core
 from pynicotine.gtkgui.application import GTK_API_VERSION
-from pynicotine.gtkgui.utils import copy_text
+from pynicotine.gtkgui.widgets import clipboard
 from pynicotine.gtkgui.widgets.accelerator import Accelerator
 from pynicotine.gtkgui.widgets.dialogs import EntryDialog
-
-
-""" Popup/Context Menu """
+from pynicotine.utils import TRANSLATE_PUNCTUATION
 
 
 class PopupMenu:
 
-    def __init__(self, frame, parent=None, callback=None, connect_events=True):
+    popup_id_counter = 0
+
+    def __init__(self, application, parent=None, callback=None, connect_events=True):
 
         self.model = Gio.Menu()
-        self.frame = frame
-        self.core = frame.core
+        self.application = application
         self.parent = parent
         self.callback = callback
 
@@ -65,10 +61,12 @@ class PopupMenu:
         self.menu_section = None
         self.editing = False
 
-        self.popup_id = ''.join(random.choice(string.digits) for _ in range(8))
+        PopupMenu.popup_id_counter += 1
+        self.popup_id = PopupMenu.popup_id_counter
 
-        self.user = None
-        self.useritem = None
+    def destroy(self):
+        self.clear()
+        self.__dict__.clear()
 
     def set_parent(self, parent):
 
@@ -91,9 +89,12 @@ class PopupMenu:
             self.popup_menu.set_parent(parent)
             self.popup_menu.set_halign(Gtk.Align.START)
             self.popup_menu.set_has_arrow(False)
+
+            # Workaround for wrong widget receiving focus after closing menu in GTK 4
+            self.popup_menu.connect("closed", lambda *_args: self.parent.child_focus(Gtk.DirectionType.TAB_FORWARD))
         else:
             self.popup_menu = Gtk.Menu.new_from_model(self.model)
-            self.popup_menu.attach_to_widget(parent, None)
+            self.popup_menu.attach_to_widget(parent)
 
         return self.popup_menu
 
@@ -102,7 +103,7 @@ class PopupMenu:
         state = GLib.Variant("b", False) if stateful else None
         action = Gio.SimpleAction(name=action_id, state=state)
 
-        self.frame.window.add_action(action)
+        self.application.add_action(action)
         return action
 
     def _create_menu_item(self, item):
@@ -112,21 +113,22 @@ class PopupMenu:
             $ - boolean
             O - choice
             # - regular
-            U - user
+            = - hidden when disabled
         """
 
         submenu = False
         boolean = False
         choice = False
+        item_type = item[0][0]
         label = item[0][1:]
 
-        if item[0][0] == ">":
+        if item_type == ">":
             submenu = True
 
-        elif item[0][0] == "$":
+        elif item_type == "$":
             boolean = True
 
-        elif item[0][0] == "O":
+        elif item_type == "O":
             choice = True
 
         if isinstance(item[1], str):
@@ -135,19 +137,18 @@ class PopupMenu:
             action = None
 
         else:
-            action_id = "win." + (label + self.popup_id).replace(" ", "").lower().translate(
-                str.maketrans(dict.fromkeys(string.punctuation)))
-
-            action = self._create_action(action_id[4:], (boolean or choice))
+            normalized_label = "-".join(label.translate(TRANSLATE_PUNCTUATION).lower().split())
+            action_id = f"app.menu-{normalized_label}-{self.popup_id}"
+            action = self._create_action(action_id.replace("app.", "", 1), (boolean or choice))
 
         if choice and len(item) > 2 and isinstance(item[2], str):
             # Choice target name
-            action_id = action_id + "::" + item[2]
+            action_id = f"{action_id}::{item[2]}"
 
         menuitem = Gio.MenuItem.new(label, action_id)
 
-        if item[0][0] == "U":
-            self.useritem = menuitem
+        if item_type == "=":
+            menuitem.set_attribute_value("hidden-when", GLib.Variant("s", "action-disabled"))
 
         if submenu:
             menuitem.set_submenu(item[1].model)
@@ -174,7 +175,7 @@ class PopupMenu:
             # Create new section
 
             self.menu_section = Gio.Menu()
-            menuitem = Gio.MenuItem.new_section(None, self.menu_section)
+            menuitem = Gio.MenuItem.new_section(label=None, section=self.menu_section)
             self.model.append_item(menuitem)
 
             if not item[0]:
@@ -184,8 +185,8 @@ class PopupMenu:
         self.menu_section.append_item(menuitem)
 
     def update_model(self):
-        """ This function is called before a menu model needs to be manipulated
-        (enabling/disabling actions, showing a menu in the GUI) """
+        """This function is called before a menu model needs to be manipulated
+        (enabling/disabling actions, showing a menu in the GUI)"""
 
         if not self.pending_items:
             return
@@ -212,15 +213,14 @@ class PopupMenu:
         self.model.remove_all()
 
         for action in self.actions.values():
-            self.frame.window.remove_action(action.get_name())
+            self.application.remove_action(action.get_name())
 
         self.actions.clear()
         self.items.clear()
 
         self.menu_section = None
-        self.useritem = None
 
-    def popup(self, pos_x, pos_y, controller=None, button=3, menu=None):
+    def popup(self, pos_x, pos_y, controller=None, menu=None):
 
         if menu is None:
             menu = self.create_context_menu(self.parent)
@@ -232,28 +232,26 @@ class PopupMenu:
             rectangle = Gdk.Rectangle()
             rectangle.x = pos_x
             rectangle.y = pos_y
-            rectangle.width = rectangle.height = 1
+
+            # Width/height 4 instead of 1 to work around this GTK bug in most cases:
+            # https://gitlab.gnome.org/GNOME/gtk/-/issues/5712
+            rectangle.width = rectangle.height = 4
 
             menu.set_pointing_to(rectangle)
             menu.popup()
             return
 
-        try:
-            event = None
+        event = None
 
-            if controller:
-                sequence = controller.get_current_sequence()
+        if controller:
+            sequence = controller.get_current_sequence()
 
-                if sequence:
-                    event = controller.get_last_event(sequence)
+            if sequence:
+                event = controller.get_last_event(sequence)
 
-            menu.popup_at_pointer(event)
+        menu.popup_at_pointer(event)
 
-        except (AttributeError, TypeError):
-            time = Gtk.get_current_event_time()
-            menu.popup(None, None, None, None, button, time)
-
-    """ Events """
+    # Events #
 
     def _callback(self, controller=None, pos_x=None, pos_y=None):
 
@@ -321,6 +319,20 @@ class PopupMenu:
 
 class FilePopupMenu(PopupMenu):
 
+    def __init__(self, application, parent=None, callback=None, connect_events=True):
+
+        super().__init__(application=application, parent=parent, callback=callback,
+                         connect_events=connect_events)
+
+        self._setup_file_menu()
+
+    def _setup_file_menu(self):
+
+        self.add_items(
+            ("#" + "selected_files", None),
+            ("", None)
+        )
+
     def set_num_selected_files(self, num_files):
 
         self.actions["selected_files"].set_enabled(False)
@@ -331,30 +343,37 @@ class FilePopupMenu(PopupMenu):
 
 class UserPopupMenu(PopupMenu):
 
-    def setup_user_menu(self, user=None, page=""):
+    def __init__(self, application, parent=None, callback=None, connect_events=True, username=None,
+                 tab_name=None):
 
-        user_label = "U"
+        super().__init__(application=application, parent=parent, callback=callback,
+                         connect_events=connect_events)
 
-        if user is not None:
-            self.user = user
-            user_label += self.user
+        self.username = username
+        self.tab_name = tab_name
+
+        self.setup_user_menu(username)
+
+    def setup_user_menu(self, username):
+
+        self.set_user(username)
 
         self.add_items(
-            (user_label, self.on_copy_user),
+            ("#" + "username", self.on_copy_user),
             ("", None)
         )
 
-        if page != "privatechat":
-            self.add_items(("#" + _("Send M_essage"), self.on_send_message))
+        if self.tab_name != "userinfo":
+            self.add_items(("#" + _("View User _Profile"), self.on_user_profile))
 
-        if page != "userinfo":
-            self.add_items(("#" + _("Show User I_nfo"), self.on_get_user_info))
+        if self.tab_name != "privatechat":
+            self.add_items(("#" + _("_Send Message"), self.on_send_message))
 
-        if page != "userbrowse":
+        if self.tab_name != "userbrowse":
             self.add_items(("#" + _("_Browse Files"), self.on_browse_user))
 
-        if page != "userlist":
-            self.add_items(("$" + _("_Add to Buddy List"), self.on_add_to_list))
+        if self.tab_name != "userlist":
+            self.add_items(("$" + _("_Add Buddy"), self.on_add_to_list))
 
         self.add_items(
             ("#" + _("_Gift Privileges…"), self.on_give_privileges),
@@ -362,44 +381,53 @@ class UserPopupMenu(PopupMenu):
             ("$" + _("Ban User"), self.on_ban_user),
             ("$" + _("Ignore User"), self.on_ignore_user),
             ("", None),
-            ("$" + _("Ban IP Address"), self.on_block_user),
+            ("$" + _("Ban IP Address"), self.on_ban_ip),
             ("$" + _("Ignore IP Address"), self.on_ignore_ip),
             ("#" + _("Show IP A_ddress"), self.on_show_ip_address)
         )
 
-    def set_user(self, user):
+    def update_username_item(self):
 
-        if not user or self.user == user:
+        if not self.username:
             return
 
-        self.user = user
+        user_item = self.items.get("username")
 
-        if not self.useritem:
+        if not user_item:
             return
 
-        self.useritem.set_label(user.replace('_', '__'))  # Escape underscores to disable mnemonics
+        user_item.set_label(self.username.replace("_", "__"))  # Escape underscores to disable mnemonics
         self.model.remove(0)
-        self.model.prepend_item(self.useritem)
+        self.model.prepend_item(user_item)
+
+    def set_user(self, username):
+
+        if username == self.username:
+            return
+
+        self.username = username
+        self.update_username_item()
 
     def toggle_user_items(self):
 
         self.editing = True
-        self.actions[_("_Gift Privileges…")].set_enabled(bool(self.core.privileges_left))
+        self.actions[_("_Gift Privileges…")].set_enabled(bool(core.users.privileges_left))
 
-        add_to_list = _("_Add to Buddy List")
+        local_username = core.users.login_username or config.sections["server"]["login"]
+        add_to_list = _("_Add Buddy")
 
         if add_to_list in self.actions:
-            self.actions[add_to_list].set_state(
-                GLib.Variant("b", self.user in (i[0] for i in config.sections["server"]["userlist"]))
-            )
+            self.actions[add_to_list].set_state(GLib.Variant("b", self.username in core.buddies.users))
 
-        self.actions[_("Ban User")].set_state(GLib.Variant("b", self.core.network_filter.is_user_banned(self.user)))
-        self.actions[_("Ignore User")].set_state(
-            GLib.Variant("b", self.core.network_filter.is_user_ignored(self.user)))
-        self.actions[_("Ban IP Address")].set_state(
-            GLib.Variant("b", self.core.network_filter.get_cached_blocked_user_ip(self.user) or False))
-        self.actions[_("Ignore IP Address")].set_state(
-            GLib.Variant("b", self.core.network_filter.get_cached_ignored_user_ip(self.user) or False))
+        for action_id, value in (
+            (_("Ban User"), core.network_filter.is_user_banned(self.username)),
+            (_("Ignore User"), core.network_filter.is_user_ignored(self.username)),
+            (_("Ban IP Address"), core.network_filter.is_user_ip_banned(self.username)),
+            (_("Ignore IP Address"), core.network_filter.is_user_ip_ignored(self.username))
+        ):
+            # Disable menu item if it's our own username and we haven't banned ourselves before
+            self.actions[action_id].set_enabled(GLib.Variant("b", self.username != local_username or value))
+            self.actions[action_id].set_state(GLib.Variant("b", value))
 
         self.editing = False
 
@@ -407,19 +435,19 @@ class UserPopupMenu(PopupMenu):
 
         popup.clear()
 
-        if self.user is None:
+        if self.username is None:
             return
 
-        popup.set_user(self.user)
+        popup.set_user(self.username)
 
-        for room, data in self.core.chatrooms.private_rooms.items():
-            is_owned = self.core.chatrooms.is_private_room_owned(room)
-            is_operator = self.core.chatrooms.is_private_room_operator(room)
+        for room, data in core.chatrooms.private_rooms.items():
+            is_owned = core.chatrooms.is_private_room_owned(room)
+            is_operator = core.chatrooms.is_private_room_operator(room)
 
             if not is_owned and not is_operator:
                 continue
 
-            if self.user in data["users"]:
+            if self.username in data["users"]:
                 popup.add_items(
                     ("#" + _("Remove from Private Room %s") % room, popup.on_private_room_remove_user, room))
             else:
@@ -428,7 +456,7 @@ class UserPopupMenu(PopupMenu):
             if not is_owned:
                 continue
 
-            if self.user in data["operators"]:
+            if self.username in data["operators"]:
                 popup.add_items(
                     ("#" + _("Remove as Operator of %s") % room, popup.on_private_room_remove_operator, room))
             else:
@@ -438,42 +466,46 @@ class UserPopupMenu(PopupMenu):
 
         popup.update_model()
 
-    """ Events """
+    def update_model(self):
+        super().update_model()
+        self.update_username_item()
+
+    # Events #
 
     def on_search_user(self, *_args):
 
-        self.frame.search_mode_action.change_state(GLib.Variant("s", "user"))
-        self.frame.user_search_entry.set_text(self.user)
-        self.frame.change_main_page(self.frame.search_page)
+        self.application.window.lookup_action("search-mode").change_state(GLib.Variant("s", "user"))
+        self.application.window.user_search_entry.set_text(self.username)
+        self.application.window.change_main_page(self.application.window.search_page)
+        GLib.idle_add(lambda: self.application.window.search_entry.grab_focus() == -1, priority=GLib.PRIORITY_HIGH_IDLE)
 
     def on_send_message(self, *_args):
-        self.core.privatechats.show_user(self.user)
-        self.frame.change_main_page(self.frame.private_page)
+        core.privatechat.show_user(self.username)
 
     def on_show_ip_address(self, *_args):
-        self.core.request_ip_address(self.user)
+        core.users.request_ip_address(self.username, notify=True)
 
-    def on_get_user_info(self, *_args):
-        self.core.userinfo.request_user_info(self.user)
+    def on_user_profile(self, *_args):
+        core.userinfo.show_user(self.username)
 
     def on_browse_user(self, *_args):
-        self.core.userbrowse.browse_user(self.user)
+        core.userbrowse.browse_user(self.username)
 
     def on_private_room_add_user(self, *args):
         room = args[-1]
-        self.core.queue.append(slskmessages.PrivateRoomAddUser(room, self.user))
+        core.chatrooms.add_user_to_private_room(room, self.username)
 
     def on_private_room_remove_user(self, *args):
         room = args[-1]
-        self.core.queue.append(slskmessages.PrivateRoomRemoveUser(room, self.user))
+        core.chatrooms.remove_user_from_private_room(room, self.username)
 
     def on_private_room_add_operator(self, *args):
         room = args[-1]
-        self.core.queue.append(slskmessages.PrivateRoomAddOperator(room, self.user))
+        core.chatrooms.add_operator_to_private_room(room, self.username)
 
     def on_private_room_remove_operator(self, *args):
         room = args[-1]
-        self.core.queue.append(slskmessages.PrivateRoomRemoveOperator(room, self.user))
+        core.chatrooms.remove_operator_from_private_room(room, self.username)
 
     def on_add_to_list(self, action, state):
 
@@ -481,9 +513,9 @@ class UserPopupMenu(PopupMenu):
             return
 
         if state.get_boolean():
-            self.core.userlist.add_user(self.user)
+            core.buddies.add_buddy(self.username)
         else:
-            self.core.userlist.remove_user(self.user)
+            core.buddies.remove_buddy(self.username)
 
         action.set_state(state)
 
@@ -493,21 +525,21 @@ class UserPopupMenu(PopupMenu):
             return
 
         if state.get_boolean():
-            self.core.network_filter.ban_user(self.user)
+            core.network_filter.ban_user(self.username)
         else:
-            self.core.network_filter.unban_user(self.user)
+            core.network_filter.unban_user(self.username)
 
         action.set_state(state)
 
-    def on_block_user(self, action, state):
+    def on_ban_ip(self, action, state):
 
         if self.editing:
             return
 
         if state.get_boolean():
-            self.core.network_filter.block_user_ip(self.user)
+            core.network_filter.ban_user_ip(self.username)
         else:
-            self.core.network_filter.unblock_user_ip(self.user)
+            core.network_filter.unban_user_ip(self.username)
 
         action.set_state(state)
 
@@ -517,9 +549,9 @@ class UserPopupMenu(PopupMenu):
             return
 
         if state.get_boolean():
-            self.core.network_filter.ignore_user_ip(self.user)
+            core.network_filter.ignore_user_ip(self.username)
         else:
-            self.core.network_filter.unignore_user_ip(self.user)
+            core.network_filter.unignore_user_ip(self.username)
 
         action.set_state(state)
 
@@ -529,14 +561,14 @@ class UserPopupMenu(PopupMenu):
             return
 
         if state.get_boolean():
-            self.core.network_filter.ignore_user(self.user)
+            core.network_filter.ignore_user(self.username)
         else:
-            self.core.network_filter.unignore_user(self.user)
+            core.network_filter.unignore_user(self.username)
 
         action.set_state(state)
 
     def on_copy_user(self, *_args):
-        copy_text(self.user)
+        clipboard.copy_text(self.username)
 
     def on_give_privileges_response(self, dialog, _response_id, _data):
 
@@ -547,29 +579,30 @@ class UserPopupMenu(PopupMenu):
 
         try:
             days = int(days)
-            self.core.request_give_privileges(self.user, days)
+            core.users.request_give_privileges(self.username, days)
 
         except ValueError:
-            self.on_give_privileges(error=_("Please enter number of days!"))
+            self.on_give_privileges(error=_("Please enter number of days."))
 
     def on_give_privileges(self, *_args, error=None):
 
-        self.core.request_check_privileges()
+        core.users.request_check_privileges()
 
-        if self.core.privileges_left is None:
+        if core.users.privileges_left is None:
             days = _("Unknown")
         else:
-            days = self.core.privileges_left // 60 // 60 // 24
+            days = core.users.privileges_left // 60 // 60 // 24
 
         message = (_("Gift days of your Soulseek privileges to user %(user)s (%(days_left)s):") %
-                   {"user": self.user, "days_left": _("%(days)s days left") % {"days": days}})
+                   {"user": self.username, "days_left": _("%(days)s days left") % {"days": days}})
 
         if error:
             message += "\n\n" + error
 
         EntryDialog(
-            parent=self.frame.window,
+            parent=self.application.window,
             title=_("Gift Privileges"),
             message=message,
+            action_button_label=_("_Give Privileges"),
             callback=self.on_give_privileges_response
-        ).show()
+        ).present()
